# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 NeonGecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest

from unittest import TestCase
from unittest.mock import Mock

from mirakuru import ProcessExitedWithError
from neon_mq_connector.consumers import SelectConsumerThread
from neon_mq_connector.utils.network_utils import dict_to_b64
from port_for import get_port
from pytest_rabbitmq.factories.executor import RabbitMqExecutor
from pytest_rabbitmq.factories.process import get_config

from neon_llm_core.llm import NeonLLM
from neon_llm_core.rmq import NeonLLMMQConnector


class NeonMockLlm(NeonLLMMQConnector):
    def __init__(self, rmq_port: int):
        config = {"MQ": {"server": "127.0.0.1", "port": rmq_port,
                         "users": {
                             "mq_handler": {"user": "neon_api_utils",
                                            "password": "Klatchat2021"},
                             "neon_llm_mock_mq": {"user": "test_llm_user",
                                                  "password": "test_llm_password"}}}}
        NeonLLMMQConnector.__init__(self, config=config)
        self._model = Mock()
        self._model.ask.return_value = "Mock response"
        self._model.get_sorted_answer_indexes.return_value = [0, 1]
        self.send_message = Mock()
        self._compose_opinion_prompt = Mock(return_value="Mock opinion prompt")

    @property
    def name(self):
        return "mock_mq"

    @property
    def model(self) -> NeonLLM:
        return self._model

    def compose_opinion_prompt(self, respondent_nick: str,
                               question: str,
                               answer: str) -> str:
        return self._compose_opinion_prompt(respondent_nick, question, answer)


@pytest.fixture(scope="class")
def rmq_instance(request, tmp_path_factory):
    config = get_config(request)
    rabbit_ctl = config["ctl"]
    rabbit_server = config["server"]
    rabbit_host = "127.0.0.1"
    rabbit_port = get_port(config["port"])
    rabbit_distribution_port = get_port(
        config["distribution_port"], [rabbit_port]
    )
    assert rabbit_distribution_port
    assert (
            rabbit_distribution_port != rabbit_port
    ), "rabbit_port and distribution_port can not be the same!"

    tmpdir = tmp_path_factory.mktemp(f"pytest-rabbitmq-{request.fixturename}")

    rabbit_plugin_path = config["plugindir"]

    rabbit_logpath = config["logsdir"]

    if not rabbit_logpath:
        rabbit_logpath = tmpdir / "logs"

    rabbit_executor = RabbitMqExecutor(
        rabbit_server,
        rabbit_host,
        rabbit_port,
        rabbit_distribution_port,
        rabbit_ctl,
        logpath=rabbit_logpath,
        path=tmpdir,
        plugin_path=rabbit_plugin_path,
        node_name=config["node"],
    )

    rabbit_executor.start()

    # Init RMQ config
    rabbit_executor.rabbitctl_output("add_user", "test_llm_user",
                                     "test_llm_password")
    rabbit_executor.rabbitctl_output("add_vhost", "/llm")
    rabbit_executor.rabbitctl_output("set_permissions", "-p", "/llm",
                                     "test_llm_user", ".*", ".*", ".*")
    request.cls.rmq_instance = rabbit_executor


@pytest.mark.usefixtures("rmq_instance")
class TestNeonLLMMQConnector(TestCase):
    mq_llm: NeonMockLlm = None
    rmq_instance: RabbitMqExecutor = None

    @classmethod
    def tearDownClass(cls):
        try:
            cls.rmq_instance.stop()
        except ProcessExitedWithError:
            pass

    def setUp(self):
        if self.mq_llm is None:
            self.mq_llm = NeonMockLlm(self.rmq_instance.port)

    def test_00_init(self):
        self.assertIn(self.mq_llm.name, self.mq_llm.service_name)
        self.assertIsInstance(self.mq_llm.ovos_config, dict)
        self.assertEqual(self.mq_llm.vhost, "/llm")
        self.assertIsNotNone(self.mq_llm.model, self.mq_llm.model)
        self.assertEqual(self.mq_llm._personas_provider.service_name,
                         self.mq_llm.name)
        self.assertTrue(self.mq_llm.async_consumers_enabled)
        self.assertEqual(self.mq_llm.consumer_thread_cls, SelectConsumerThread)
        for consumer in self.mq_llm.consumers.values():
            self.assertIsInstance(consumer, SelectConsumerThread)

    def test_handle_request(self):
        from neon_data_models.models.api.mq import (LLMProposeRequest,
                                                    LLMProposeResponse)
        # Valid Request
        request = LLMProposeRequest(message_id="mock_message_id",
                                    routing_key="mock_routing_key",
                                    query="Mock Query", history=[])
        self.mq_llm.handle_request(None, None, None,
                                   dict_to_b64(request.model_dump())).join()
        self.mq_llm.model.ask.assert_called_with(message=request.query,
                                                 chat_history=request.history,
                                                 persona=request.persona)
        response = self.mq_llm.send_message.call_args.kwargs
        self.assertEqual(response['queue'], request.routing_key)
        response = LLMProposeResponse(**response['request_data'])
        self.assertIsInstance(response, LLMProposeResponse)
        self.assertEqual(request.routing_key, response.routing_key)
        self.assertEqual(request.message_id, response.message_id)

        self.assertEqual(response.response, self.mq_llm.model.ask())

    def test_handle_opinion_request(self):
        from neon_data_models.models.api.mq import (LLMDiscussRequest,
                                                    LLMDiscussResponse)
        # Valid Request
        request = LLMDiscussRequest(message_id="mock_message_id",
                                    routing_key="mock_routing_key",
                                    query="Mock Discuss", history=[],
                                    options={"bot 1": "resp 1",
                                             "bot 2": "resp 2"})
        self.mq_llm.handle_opinion_request(None, None, None,
                                           dict_to_b64(request.model_dump())).join()

        self.mq_llm._compose_opinion_prompt.assert_called_with(
            list(request.options.keys())[0], request.query,
            list(request.options.values())[0])

        response = self.mq_llm.send_message.call_args.kwargs
        self.assertEqual(response['queue'], request.routing_key)
        response = LLMDiscussResponse(**response['request_data'])
        self.assertIsInstance(response, LLMDiscussResponse)
        self.assertEqual(request.routing_key, response.routing_key)
        self.assertEqual(request.message_id, response.message_id)

        self.assertEqual(response.opinion, self.mq_llm.model.ask())

        # No input options
        request = LLMDiscussRequest(message_id="mock_message_id1",
                                    routing_key="mock_routing_key1",
                                    query="Mock Discuss 1", history=[],
                                    options={})
        self.mq_llm.handle_opinion_request(None, None, None,
                                           dict_to_b64(request.model_dump())).join()
        response = self.mq_llm.send_message.call_args.kwargs
        self.assertEqual(response['queue'], request.routing_key)
        response = LLMDiscussResponse(**response['request_data'])
        self.assertIsInstance(response, LLMDiscussResponse)
        self.assertEqual(request.routing_key, response.routing_key)
        self.assertEqual(request.message_id, response.message_id)
        self.assertNotEqual(response.opinion, self.mq_llm.model.ask())

        # TODO: Test with invalid sorted answer indexes

    def test_handle_score_request(self):
        from neon_data_models.models.api.mq import (LLMVoteRequest,
                                                    LLMVoteResponse)

        # Valid Request
        request = LLMVoteRequest(message_id="mock_message_id",
                                 routing_key="mock_routing_key",
                                 query="Mock Score", history=[],
                                 responses=["one", "two"])
        self.mq_llm.handle_score_request(None, None, None,
                                         dict_to_b64(request.model_dump())).join()

        response = self.mq_llm.send_message.call_args.kwargs
        self.assertEqual(response['queue'], request.routing_key)
        response = LLMVoteResponse(**response['request_data'])
        self.assertIsInstance(response, LLMVoteResponse)
        self.assertEqual(request.routing_key, response.routing_key)
        self.assertEqual(request.message_id, response.message_id)

        self.assertEqual(response.sorted_answer_indexes,
                         self.mq_llm.model.get_sorted_answer_indexes())

        # No response options
        request = LLMVoteRequest(message_id="mock_message_id",
                                 routing_key="mock_routing_key",
                                 query="Mock Score", history=[], responses=[])
        self.mq_llm.handle_score_request(None, None, None,
                                         dict_to_b64(request.model_dump())).join()

        response = self.mq_llm.send_message.call_args.kwargs
        self.assertEqual(response['queue'], request.routing_key)
        response = LLMVoteResponse(**response['request_data'])
        self.assertIsInstance(response, LLMVoteResponse)
        self.assertEqual(request.routing_key, response.routing_key)
        self.assertEqual(request.message_id, response.message_id)

        self.assertEqual(response.sorted_answer_indexes, [])



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 Neongecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from abc import ABC, abstractmethod
from typing import List


class NeonLLM(ABC):

    mq_to_llm_role = {}

    def __init__(self, config: dict):
        """
        @param config: Dict LLM configuration for this specific LLM
        """
        self._llm_config = config
        self._tokenizer = None
        self._model = None

    @property
    def llm_config(self):
        """
        Get the configuration for this LLM instance
        """
        return self._llm_config

    @property
    @abstractmethod
    def tokenizer(self):
        pass

    @property
    @abstractmethod
    def tokenizer_model_name(self) -> str:
        pass

    @property
    @abstractmethod
    def model(self):
        pass

    @property
    @abstractmethod
    def llm_model_name(self) -> str:
        pass

    @property
    @abstractmethod
    def _system_prompt(self) -> str:
        pass

    def ask(self, message: str, chat_history: List[List[str]], persona: dict) -> str:
        """ Generates llm response based on user message and (user, llm) chat history """
        prompt = self._assemble_prompt(message, chat_history, persona)
        llm_text_output = self._call_model(prompt)
        return llm_text_output

    @abstractmethod
    def get_sorted_answer_indexes(self, question: str, answers: List[str], persona: dict) -> List[int]:
        """
            Creates sorted list of answer indexes with respect to order provided in :param answers
            Results should be sorted from best to worst
            :param question: incoming question
            :param answers: list of answers to rank
            :returns list of indexes
        """
        pass

    @abstractmethod
    def _call_model(self, prompt: str) -> str:
        """
        Wrapper for Model generation logic. This method may be called
        asynchronously, so it is up to the extending class to use locks or
        queue inputs as necessary.
        :param prompt: Input text sequence
        :returns: Output text sequence generated by model
        """
        pass

    @abstractmethod
    def _assemble_prompt(self, message: str, chat_history: List[List[str]], persona: dict):
        """
            Assembles prompt engineering logic

            :param message: Incoming prompt
            :param chat_history: History of preceding conversation
            :returns: assembled prompt
        """
        pass

    @abstractmethod
    def _tokenize(self, prompt: str) -> List[str]:
        pass

    @classmethod
    def convert_role(cls, role: str) -> str:
        """ Maps MQ role to LLM's internal domain """
        matching_llm_role = cls.mq_to_llm_role.get(role)
        if not matching_llm_role:
            raise ValueError(f"role={role} is undefined, supported are: "
                             f"{list(cls.mq_to_llm_role)}")
        return matching_llm_role



# NEON AI (TM) SOFTWARE, Software Development Kit & Application Development System
# All trademark and other rights reserved by their respective owners
# Copyright 2008-2025 NeonGecko.com Inc.
# BSD-3
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS;  OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE,  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import os
from time import time
from typing import List, Optional

from neon_data_models.models.api.llm import (
    LLMPersona,
    LLMPersonaIdentity,
)

from neon_mq_connector.utils import RepeatingTimer
from neon_mq_connector.utils.client_utils import send_mq_request
from neon_utils.logger import LOG
from pydantic import ValidationError

from neon_llm_core.utils.constants import LLM_VHOST
from neon_llm_core.utils.personas.state import PersonaHandlersState


class PersonasProvider:
    """
    Manages personas defined via Klat. Each LLM that connects to the MQ bus will
    include an instance of this object to track changes to personas.
    """

    PERSONA_STATE_TTL = int(os.getenv("PERSONA_STATE_TTL", 15 * 60))
    PERSONA_SYNC_INTERVAL = int(os.getenv("PERSONA_SYNC_INTERVAL", 0))
    GET_CONFIGURED_PERSONAS_QUEUE = "get_configured_personas"

    def __init__(self, service_name: str, ovos_config: dict):
        self.service_name = service_name
        self._persona_handlers_state = PersonaHandlersState(service_name=service_name,
                                                            ovos_config=ovos_config)
        self._personas = []  # list of personas available for given service
        self._persona_last_sync = 0
        self._persona_sync_thread = None

    @property
    def persona_sync_thread(self) -> RepeatingTimer:
        """Creates new synchronization thread which fetches Klat personas"""
        if not (isinstance(self._persona_sync_thread, RepeatingTimer) and
                self._persona_sync_thread.is_alive()):
            self._persona_sync_thread = RepeatingTimer(self.PERSONA_SYNC_INTERVAL,
                                                       self._fetch_persona_config)
            self._persona_sync_thread.daemon = True
        return self._persona_sync_thread

    @property
    def personas(self) -> List[LLMPersona]:
        return self._personas

    @personas.setter
    def personas(self, data: List[LLMPersona]):
        LOG.debug(f'Setting personas={data}')
        if self._should_reset_personas(data=data):
            LOG.warning(f'Persona state expired, setting default personas')
            self._personas = []
            self._persona_handlers_state.init_default_personas()
        else:
            self._personas = data
            self._persona_handlers_state.clean_up_personas(ignore_items=self._personas)

    def _should_reset_personas(self, data: List[LLMPersona]) -> bool:
        """
        Checks if personas should be re-initialized after setting a new value
        for personas.

        If PERSONA_SYNC_INTERVAL is enabled - verifies based on TTL, otherwise

        :param data: requested list of personas
        :return: True if requested `data` should be ignored and personas
            reloaded from config, False if requested `data` should be used
            directly
        """
        return self._should_reset_personas_based_on_ttl(data) if self.PERSONA_SYNC_INTERVAL > 0 else not data

    def _should_reset_personas_based_on_ttl(self, data: dict) -> bool:
        """
        Determines whether personas should be reset based on Time-to-Live (TTL) and
        the synchronization timestamp.

        Examines the time elapsed since the last persona synchronization in relation
        to a predefined TTL. Also considers whether the state of the personas and the
        incoming data indicate a need for resetting.

        :param data: provided persona data

        returns: True if personas need to be reset based on TTL False otherwise.
        """
        return (not (self._persona_last_sync == 0 and data)
                and int(time()) - self._persona_last_sync > self.PERSONA_STATE_TTL)

    def _fetch_persona_config(self):
        """
        Get personas from a provider on the MQ bus and update the internal
        `personas` reference.
        """
        response = send_mq_request(
            vhost=LLM_VHOST,
            request_data={"service_name": self.service_name},
            target_queue=PersonasProvider.GET_CONFIGURED_PERSONAS_QUEUE,
            timeout=60)
        self.parse_persona_config_response(response)

    def parse_persona_config_response(self, persona_response: dict):
        """
        Parses and processes a response containing persona data, updates internal state,
        and manages personas accordingly.

        :param persona_response: A dictionary containing the response data with
                                 persona information.
                                 Expected to contain a key 'items' holding a list of
                                 persona details.
        """
        if 'items' in persona_response:
            self._persona_last_sync = int(time())
        response_data = persona_response.get('items', [])
        active_personas = []
        for persona_data in response_data:
            applied_persona = self.apply_persona_data(persona_data=persona_data)
            if applied_persona:
                active_personas.append(applied_persona)
        self.personas = active_personas

    def apply_persona_data(self, persona_data: dict) -> Optional[LLMPersona]:
        """
        Validates and applies persona data.

        Takes a dictionary representing persona data, validates it,
        and applies the validated persona to the current instance.
        If the validation succeeds, it returns the applied persona. Otherwise, returns None.

        :param persona_data: Dictionary containing persona data

        :returns: The validated and applied persona instance if validation is successful, or None otherwise.
        """
        validated_persona = self._validate_persona_data(persona_data)
        if validated_persona:
            persona_added = self._add_persona(persona=validated_persona)
            if persona_added:
                return validated_persona

    @staticmethod
    def _validate_persona_data(persona_data: dict) -> Optional[LLMPersona]:
        """
        Validates of the persona data and returns the resulting `PersonaModel`.
        If validation fails - logs error and returns None

        :param persona_data : A dictionary containing details of the persona, where
                              specific key-value mappings are applied for validation.

        returns: A validated and updated `PersonaModel` instance based on the provided data if
                 validation was successful, None otherwise
        """
        persona_data.setdefault('name', persona_data.pop('persona_name', None))

        try:
            return LLMPersona.model_validate(obj=persona_data)
        except ValidationError as err:
            LOG.error(f"Failed to apply persona data from {persona_data} - {str(err)}")
            return

    @staticmethod
    def _validate_persona_identity(persona_identity_data: dict) -> Optional[LLMPersonaIdentity]:
        """
        Validates of the persona data and returns the resulting `PersonaModel`.
        If validation fails - logs error and returns None

        :param persona_identity_data : A dictionary containing details of the persona identity, where
                                       specific key-value mappings are applied for validation.

        returns: A validated and updated `LLMPersonaIdentity` instance based on the provided data if
                 validation was successful, None otherwise
        """
        persona_identity_data.setdefault('name', persona_identity_data.pop('persona_name', None))

        try:
            return LLMPersonaIdentity.model_validate(obj=persona_identity_data)
        except ValidationError as err:
            LOG.error(f"Failed to apply persona identity data from {persona_identity_data} - {str(err)}")
            return

    def _add_persona(self, persona: LLMPersona) -> bool:
        """
        Attempts to add incoming persona and return an updated PersonaModel instance if successful.
        If default personas are running upon adding persona - removes default personas.
        If state container has only one running persona & it was disabled by this method -
        triggers initialisation of the default personas.

        :param persona: `LLMPersona` instance to add to the `PersonaHandlersState` container

        :returns: True if incoming persona object in case of successful addition or if identical persona already exists
                  False otherwise
        """

        new_persona = self._persona_handlers_state.add_persona_handler(persona=persona)

        if new_persona:
            LOG.info(f"Persona {persona.id} updated successfully")

            # Once first manually configured persona added - pruning default personas
            if self._persona_handlers_state.default_personas_running:
                LOG.info("Starting to remove default personas")
                self._persona_handlers_state.clean_up_personas(ignore_items=[persona])
                self._persona_handlers_state.default_personas_running = False
                LOG.info("Completed removing of default personas")

        elif persona.id not in self._persona_handlers_state.connected_persona_ids:
            # May occur if the last updated persona was set to be disabled
            if not self._persona_handlers_state.has_connected_personas():
                LOG.info("No personas connected after the last update - setting default personas")
                self._persona_handlers_state.init_default_personas()
            return False

        return True

    def remove_persona(self, persona_data: dict):
        """
        Removes a persona from the active persona handlers state.

        This method handles the removal of a persona based on persona data.
        It ensures that the default personas are initialized if no other connected personas
        remain after the removal.

        :param persona_data: A dictionary containing details of the persona to be removed.
        """
        if (self._persona_handlers_state.has_connected_personas() and
                not self._persona_handlers_state.default_personas_running):
            persona_identity = self._validate_persona_identity(persona_identity_data=persona_data)
            if persona_identity:
                self._persona_handlers_state.remove_persona(persona_id=persona_identity.id)
                LOG.info(f"Persona {persona_identity.id} removed successfully")
                if not self._persona_handlers_state.has_connected_personas():
                    LOG.info("No personas connected after the last removal - setting default personas")
                    self._persona_handlers_state.init_default_personas()
        else:
            LOG.warning("No running personas detected - skipping persona removal")

    def start_sync(self):
        """
        Update personas and start thread to periodically update from a service
        on the MQ bus.
        """
        self._fetch_persona_config()
        if self.PERSONA_SYNC_INTERVAL > 0:
            self.persona_sync_thread.start()

    def stop_sync(self):
        """
        Stop persona updates from the MQ bus.
        """
        if self._persona_sync_thread:
            self._persona_sync_thread.cancel()
            self._persona_sync_thread = None
